# Copyright (c) 2020 Andrii Shekhovtsov

import numpy as np

__all__ = [
    'minmax_normalization',
    'max_normalization',
    'sum_normalization',
    'vector_normalization',
    'logaritmic_normalization',
    'normalize_matrix'
]

def minmax_normalization(x, cost=False):
    if np.min(x) == np.max(x): # If all values are equal
        return np.ones(x.shape)

    if cost:
        return (np.max(x) - x) / (np.max(x) - np.min(x))
    return (x - np.min(x)) / (np.max(x) - np.min(x))


def max_normalization(x, cost=False):
    if cost:
        return 1 - x/np.max(x)
    return x / np.max(x)


def sum_normalization(x, cost=False):
    if cost:
        return (1/x) / np.sum(1/x)
    return x / np.sum(x)


def vector_normalization(x, cost=False):
    if cost:
        return 1 - (x / np.sqrt(sum(x ** 2)))
    return x / np.sqrt(np.sum(x ** 2))


def logaritmic_normalization(x, cost=False):
    prod = np.prod(x)
    if cost:
        return (1 - (np.log(x) / np.log(prod))) / (x.shape[0] - 1)
    return np.log(x) / np.log(prod)


def normalize_matrix(matrix, method, criteria_types):
    """Normalize each column in `matrix`, using `normalization` function according `criteria_types`.

Parameters
----------
    matrix : ndarray
        Decision matrix representation.
        The rows are considered as alternatives and the columns are considered as criteria.

    normalization : callable
        Function which should be used for normalize `matrix` columns.
        It should match signature `foo(x, cost)`, where `x` is a vector which would be normalized and `cost` is a bool variable which says if `x` is a cost or profit criteria.

    criteria_types : None or ndarray
        Describes criteria types.
        1 if criteria is profit and -1 if criteria is cost for each criteria in `matrix`.
        If None all criteria are considered as profit

Returns
-------
    ndarray
        Normalized copy of the input matrix.

Raises
------
    ValueError
        If `criteria_types` and `matrix` has different number of criteria.
"""
    if criteria_types is None:
        nmatrix = matrix.astype('float')
        for i in range(matrix.shape[1]):
            nmatrix[:,i] = method(matrix[:,i], cost=False)
        return nmatrix

    if matrix.shape[1] != len(criteria_types):
        raise ValueError(f'Matrix has {matrix.shape[1]} criteria and criteria_types has {len(criteria_types)}. This values must be equal.')

    nmatrix = matrix.astype('float')
    for i, type in enumerate(criteria_types):
        if type == 1: # If profit
            nmatrix[:,i] = method(matrix[:,i], cost=False)
        else:
            nmatrix[:,i] = method(matrix[:,i], cost=True)
    return nmatrix




# Copyright (c) 2020 Andrii Shekhovtsov

import numpy as np
from .mcda_method import MCDA_method

class COPRAS(MCDA_method):
    def __init__(self):
        """Create COPRAS method object."""
        pass

    def __call__(self, matrix, weights, types, *args, **kwargs):
        """Rank alternatives from decision matrix `matrix`, with criteria weights `weights` and criteria types `types`.

Parameters
----------
    matrix : ndarray
        Decision matrix / alternatives data.
        Alternatives are in rows and Criteria are in columns.

    weights : ndarray
        Criteria weights. Sum of the weights should be 1. (e.g. sum(weights) == 1)

    types : ndarray
        Array with definitions of criteria types:
        1 if criteria is profit and -1 if criteria is cost for each criteria in `matrix`.

    *args and **kwargs are necessary for methods which reqiure some additional data.

Returns
-------
    ndarray
        Preference values for alternatives. Better alternatives have higher values.
"""
        COPRAS._validate_input_data(matrix, weights, types)
        if np.all(types == 1.0):
            raise ValueError('types array contains only profit criteria. COPRAS method requires at least one cost criteria.')
        return COPRAS._copras(matrix, weights, types)

    @staticmethod
    def _copras(matrix, weights, cryteria_types):
        nmatrix = matrix / np.sum(matrix, axis=0)

        # Difficult normalized decision making matrix
        wmatrix = nmatrix * weights

        Sp = np.sum(wmatrix[:, cryteria_types == 1], axis=1)
        Sm = np.sum(wmatrix[:, cryteria_types == -1], axis=1)

        Q = Sp + ((np.min(Sm) * np.sum(Sm))\
                / (Sm * np.sum(np.min(Sm) / Sm)))

        return Q / np.max(Q)



# Copyright (c) 2020 Andrii Shekhovtsov

from itertools import product
from functools import reduce

import numpy as np

from .mcda_method import MCDA_method
from .topsis import TOPSIS


def _TFN(a, m, b):
    def tfn(x):
        res = np.zeros(x.shape)
        mask = x == m
        res[mask] = 1

        mask = np.logical_and(x > a, x < m)
        res[mask] = (x[mask] - a) / (m - a)

        mask = np.logical_and(x < b, x > m)
        res[mask] = (b - x[mask]) / (b - m)
        return res
    return tfn


class COMET(MCDA_method):
    def __init__(self, cvalues, rate_function=None, expert_function=None):
        """Initialize COMET model. It creates CO and rank them using `ranking_method`.

Parameters
----------
    cvalues : ndarray or list of lists
        Each row represent characteristic values for each criteria.

    rate_function : callable
        Function to rate CO without creating MEJ. Matrix with CO as rows is passed as an argument Vector with rates should be retrurn. Better CO should has higher values.

        Signature of the function should be as followed:
            rate_function(co: np.array) -> np.array

    expert_function : callable
        Function which would be used to compare CO on MEJ creation.
        It should fulfill this requirments:
           CO to compare are passed as arguments (a, b)
           if a is better then b return 1,
           if b is better then a return 0,
           if this CO are equaly prefered return 0.5

    If both ranking_method and expert_function are provided, expert_function is preffered.
"""
        # Validate input
        for i, cv in enumerate(cvalues):
            if len(cv) < 2:
                raise ValueError(
                    f'You should provide minimum 2 characteristic value for each criterion. Check criterion with index {i}.'
                    )
            # Check if sorted
            if any(cv[i] >= cv[i+1] for i in range(len(cv)-1)):
                raise ValueError(
                        f'Characteristic values must be sorted in ascending order and does not contain repeated elements. Check criterion with index {i}.'
                    )

        co = product(*cvalues)
        co = np.array(list(co))

        # Determine how MEJ and SJ is calculated
        if expert_function is not None:
            self.mej = COMET._build_mej(co, expert_function)
            sj = np.sum(self.mej , axis=1)
        elif rate_function is not None:
            self.mej = None
            sj = rate_function(co)
            if sj.shape[0] != co.shape[0]:
                raise ValueError(
                        f'Rate function must returns vector with same length as number of characteristic objects. Expected length: {co.shape[0]}, but returned vector has length {sj.shape[0]}.'
                    )
        else:
            raise ValueError('rate_function or expert_function should be provided.')

        k = np.unique(sj).shape[0]

        p = np.zeros(sj.shape[0], dtype=float)
        for i in range(1, k):
            ind = sj == np.max(sj)
            p[ind] = (k - i) / (k - 1)
            sj[ind] = 0

        self.criterion_number = len(cvalues)
        self.cvalues = cvalues
        self.p = p
        self.tfns = [COMET._make_tfns(chv) for chv in cvalues]


    def __call__(self, alts, *args, **kwargs):
        """Rank alternatives from decision matrix `alts`, with criteria weights `weights` and criteria types `types`.

Parameters
----------
    alts : ndarray
        Decision matrix / alternatives data.
        Alternatives are in rows and Criteria are in columns.

    *args and **kwargs are necessary for methods which reqiure some additional data.

Returns
-------
    ndarray
        Preference values for alternatives. Better alternatives have higher values.
"""
        if self.criterion_number != alts.shape[1]:
            raise ValueError(
                    'Number of criteria in decision matrix must be equal to number of criteria in characteristic values.'
                )

        tfns = self.tfns

        pref_level_vectors = [[tfn(values) for tfn in tfns_icrit]
                              for values, tfns_icrit in zip(alts.T, tfns)]

        tfns_values_product = product(*pref_level_vectors)
        multiplayed_co = (reduce(lambda a, b: a*b, co_values) * p
                          for p, co_values in zip(self.p, tfns_values_product))
        return sum(multiplayed_co)


    @staticmethod
    def _build_mej(co, expert_function):
        # Initiate MEJ with diagonal with 0.5 values
        mej = np.diag(np.ones(co.shape[0]) * 0.5)
        for i in range(mej.shape[0]):
            for j in range(i+1, mej.shape[0]):
                v = expert_function(co[i], co[j])
                mej[i, j] = v
                mej[j, i] = 1 - v
        return mej

    def get_MEJ(self):
        if self.mej is not None:
            return self.mej

        # If there's no MEJ then rate_function was used to create SJ and P
        # Now we can create MEJ using P to compare CO.
        p = self.p
        lenp = len(p)
        mej = np.diag(np.ones(lenp)/2)
        for i in range(lenp):
            for j in range(i+1, lenp):
                if p[i] < p[j]:
                    mej[i, j] = 0.0
                    mej[j, i] = 1.0
                elif p[i] == p[j]:
                    mej[i, j] = 0.5
                    mej[j, i] = 0.5
                else:
                    mej[i, j] = 1.0
                    mej[j, i] = 0.0
        self.mej = mej
        return mej

    @staticmethod
    def _make_tfns(chv):
        tfns = []
        # First TFN
        tfns.append(_TFN(chv[0], chv[0], chv[1]))

        for i in range(1, len(chv)-1):
            tfns.append(_TFN(chv[i-1], chv[i], chv[i+1]))

        # Last TFN
        tfns.append(_TFN(chv[-2], chv[-1], chv[-1]))

        return tfns

    @staticmethod
    def manual_expert(criteria_names):
        """Returns function for manual rating characteristic objects."""
        def manual(a, b):
            # Print CO data
            print(f'{" "*15} | {"a":>9} | {"b":>9}')
            for name, va, vb in zip(criteria_names, a, b):
                print(f'{name[:15]:>15s} | {va:>9} | {vb:>9}')

            # Input from expert
            print('Which one is better?')
            options = {'a': 1.0, 'b': 0.0, '': 0.5}
            inp = None
            while inp not in options:
                inp = input('Input "a", "b" or leave empty if a i b are equaly prefered.\n>>> ')
            return options[inp]
        return manual

    @staticmethod
    def topsis_rate_function(weights, types):
        """Returns function to rate characteristic objects with TOPSIS"""
        topsis = TOPSIS()
        def topsis_rate(co):
            return topsis(co, weights, types)
        return topsis_rate




# Copyright (c) 2020 Andrii Shekhovtsov

import numpy as np
from .. import normalizations
from .mcda_method import MCDA_method


class TOPSIS(MCDA_method):
    def __init__(self, normalization_function=normalizations.minmax_normalization):
        """Create TOPSIS method object, using normaliztion `normalization_function`.

Parameters
----------
    normalization_function : callable
        Function which should be used to normalize `matrix` columns. It should match signature `foo(x, cost)`, where `x` is a vector which should be normalized and `cost` is a bool variable which says if `x` is a cost or profit criterion.
"""
        self.normalization = normalization_function

    def __call__(self, matrix, weights, types, *args, **kwargs):
        """Rank alternatives from decision matrix `matrix`, with criteria weights `weights` and criteria types `types`.

Parameters
----------
    matrix : ndarray
        Decision matrix / alternatives data.
        Alternatives are in rows and Criteria are in columns.

    weights : ndarray
        Criteria weights. Sum of the weights should be 1. (e.g. sum(weights) == 1)

    types : ndarray
        Array with definitions of criteria types:
        1 if criteria is profit and -1 if criteria is cost for each criteria in `matrix`.

    *args and **kwargs are necessary for methods which reqiure some additional data.

Returns
-------
    ndarray
        Preference values for alternatives. Better alternatives have higher values.
"""
        TOPSIS._validate_input_data(matrix, weights, types)
        if self.normalization is not None:
            nmatrix = normalizations.normalize_matrix(matrix, self.normalization, types)
        else:
            nmatrix = normalizations.normalize_matrix(matrix, normalization.minmax_normalization, types)
        return TOPSIS._topsis(nmatrix, weights)

    @staticmethod
    def _topsis(nmatrix, weights):
        # Every row of nmatrix is multiplayed by weights
        weighted_matrix = nmatrix * weights

        # Vectors of PIS and NIS
        pis = np.max(weighted_matrix, axis=0)
        nis = np.min(weighted_matrix, axis=0)

        # PIS and NIS are substracted from every row of weighted matrix
        Dp = np.sqrt(np.sum((weighted_matrix - pis) ** 2, axis=1))
        Dm = np.sqrt(np.sum((weighted_matrix - nis) ** 2, axis=1))

        return Dm / (Dm + Dp)




# Copyright (c) 2020 Andrii Shekhovtsov

import numpy as np
from .. import normalizations
from .mcda_method import MCDA_method

def _fake_normalization(x, cost=False):
    if cost:
        return np.max(x) - x
    else:
        return x


class VIKOR(MCDA_method):
    def __init__(self, normalization_function=None):
        """Create VIKOR method object, using normaliztion `normalization_function`.

Parameters
----------
    normalization_function : None or callable
        Function which should be used to normalize `matrix` columns. It should match signature `foo(x, cost)`, where `x` is a vector which should be normalized and `cost` is a bool variable which says if `x` is a cost or profit criterion.
"""
        if normalization_function is None:
            self.normalization = _fake_normalization
        else:
            self.normalization = normalization_function

    def __call__(self, matrix, weights, types, *args, v=0.5, return_all=False, **kwargs):
        """Rank alternatives from decision matrix `matrix`, with criteria weights `weights` and criteria types `types`.

Parameters
----------
    matrix : ndarray
        Decision matrix / alternatives data.
        Alternatives are in rows and Criteria are in columns.

    weights : ndarray
        Criteria weights. Sum of the weights should be 1. (e.g. sum(weights) == 1)

    types : ndarray
        Array with definitions of criteria types:
        1 if criteria is profit and -1 if criteria is cost for each criteria in `matrix`.

    v : float
        Weight of the strategy (see VIKOR algorithm explanation).

    return_all : bool
        If True, all three ranking (S, R, Q) would be returned.

    *args and **kwargs are necessary for methods which reqiure some additional data.

Returns
-------
    if `return_all` is False
    ndarray
        Q preference values for alternatives. Better alternatives have smaller values.

    if `reeturn_all` is True
    ndarray, ndarray, ndarray
        S, R, Q preference values (see VIKOR algorithm explanation).
"""
        VIKOR._validate_input_data(matrix, weights, types)
        nmatrix = normalizations.normalize_matrix(matrix, self.normalization, types)
        S, R, Q = VIKOR._vikor(nmatrix, weights, v)
        if return_all:
            return S, R, Q
        else:
            return Q

    @staticmethod
    def _vikor(matrix, weights, v=0.5):
        """
VIKOR MCDM method

Arguments:
    matrix: Decision matrix.
            Alternative are in rows and Criteria are in columns.
    weights: Weights to criteria
Returns:
    S, R, Q: Ranking lists
"""
        fstar = np.max(matrix, axis=0)
        fminus = np.min(matrix, axis=0)

        if np.any(fstar == fminus):
            eq = np.arange(fstar.shape[0])[fstar == fminus]
            raise ValueError(
                f'Criteria with indexes {eq} contains equal values for all alternatives. VIKOR method could not be applied in this case. Consider removing this criteria from the decision matrix or use another MCDA method.'
            )

        weighted_ff = weights * ((fstar - matrix)/(fstar - fminus))
        S = np.sum(weighted_ff, axis=1)
        R = np.max(weighted_ff, axis=1)

        Sstar = np.min(S)
        Sminus = np.max(S)
        Rstar = np.min(R)
        Rminus = np.max(R)

        Q = v * (S - Sstar)/(Sminus - Sstar)\
          + (1 - v) * (R - Rstar)/(Rminus - Rstar)

        return S, R, Q



# Copyright (c) 2020 Andrii Shekhovtsov

from functools import partial
import numpy as np

from .. import normalizations
from .mcda_method import MCDA_method


class PROMETHEE_II(MCDA_method):
    def __init__(self, preference_function):
        """Create PROMEHTEE_II method object, with `preference_function`.

Parameters
----------
    preference_function: str
        Name of the preference function ('usual', 'ushape', 'vshape', 'level', 'vshape_2')
"""
        self.pf = getattr(PROMETHEE_II._PreferenceFunctions, preference_function)

    def __call__(self, matrix, weights, types, *args, p=None, q=None, promethee_I=False, **kwargs):
        """
Rank alternatives from decision matrix `matrix`, with criteria weights `weights` and criteria types `types`.

Parameters
----------
    matrix : ndarray
        Decision matrix / alternatives data.
        Alternatives are in rows and Criteria are in columns.

    weights : ndarray
        Criteria weights. Sum of the weights should be 1. (e.g. sum(weights) == 1)

    types : ndarray
        Array with definitions of criteria types:
        1 if criteria is profit and -1 if criteria is cost for each criteria in `matrix`.

    p : ndarray
        p values for each criterion

    q : ndarray
        q values for each criterion

    promethee_I : bool
        If True then returns F+ and F- (like in promethee I).

    *args and **kwargs are necessary for methods which reqiure some additional data.

Returns
-------
    If `promethee_I` is True:
    ndarray
        Positive flow

    ndarray
        Negative flow

    If `promethee_I` is False:
    ndarray
        Preference values of alternatives. Better alternatives have higher values.
"""
        pf = self.pf
        if p is None and q is None:
            pfs = (partial(pf, p=None, q=None) for i in range(matrix.shape[1]))
        elif p is None and q is not None:
            pfs = (partial(pf, p=None, q=q_) for q_ in q)
        elif p is not None and q is None:
            pfs = (partial(pf, p=p_, q=None) for p_ in p)
        else:
            pfs = (partial(pf, p=p_, q=q_) for p_, q_ in zip(p, q))

        Fp, Fm, FI = PROMETHEE_II._promethee(matrix, weights, types, pfs)
        if promethee_I:
            return Fp, Fm
        else:
            return FI

    @staticmethod
    def _promethee(matrix, weights, criteria_types, pref_functions):
        # N - number of alternatives
        # M - number of criteria
        N, M = matrix.shape

        c_tables = (np.tile(crit.reshape(N, 1), (1, N)) for crit in matrix.T)
        diff_tables = ((c - crit if ct == 1 else crit - c)
                       for crit, c, ct in zip(matrix.T, c_tables, criteria_types))

        pi_table = sum(w * pf(d) for w, d, pf in zip(weights, diff_tables, pref_functions))

        F_plus = np.sum(pi_table, axis=1) / (N-1)
        F_minus = np.sum(pi_table, axis=0) / (N-1)

        FI = F_plus - F_minus

        return F_plus, F_minus, FI


    class _PreferenceFunctions:
        @staticmethod
        def usual(d, q, p):
            return (d > 0).astype(np.int8)

        @staticmethod
        def ushape(d, q, p):
            return (d > q).astype(np.int8)

        @staticmethod
        def vshape(d, q, p):
            d_ = d.copy()
            cond = np.logical_and(0 < d, d <= p)
            np.putmask(d_, cond, d/p)
            np.putmask(d_, np.logical_not(cond), d > p)
            return d_

        @staticmethod
        def level(d, q, p):
            d_ = d.copy()
            cond = np.logical_and(q < d, d <= p)
            np.putmask(d_, cond, 0.5)
            np.putmask(d_, np.logical_not(cond), d > p)
            return d_

        @staticmethod
        def vshape_2(d, q, p):
            d_ = d.copy()
            cond = np.logical_and(q < d, d <= p)
            np.putmask(d_, cond, (d-q)/(p-q))
            np.putmask(d_, np.logical_not(cond), d > p)
            return d_



# Copyright (c) 2021 Andrii Shekhovtsov

import numpy as np
from collections import Counter

__all__ = [
    'rankdata',
    'rrankdata'
]

def rankdata(a, reverse=False):
    """
    Assign ranks to data in vector `a`.

    Ranks begin at 1. Tied elements get average rank (see Examples below).

    Ranking starts from smaller values, e.g. the smaller element get
    the first position. The `reverse` argument reverse posisions, e.g.
    the largest element get first position.

    Parameters
    ----------
    a : iterable
        The array of values to be ranked.

    reverse : bool, optional
        If True, larger elements get first posisions in ranking.
        If False, smaller elements get first positions in ranking.

    Returns
    -------
    ndarray
        An array of rank scores for the input data.

    Examples
    --------
    >>> from pymcdm.helpers import rankdata
    >>> rankdata([0, 3, 2, 5])
    array([1, 3, 2, 4])
    >>> rankdata([0, 3, 2, 5], reverse=True)
    array([4, 2, 3, 1])
    >>> rankdata([0, 3, 2, 3])
    array([1. , 3.5, 2. , 3.5])
    >>> rankdata([0, 3, 2, 3], reverse=True)
    array([4. , 1.5, 3. , 1.5])
    """
    c = Counter(a)
    rv = {}
    i = 1
    for k in sorted(c.keys(), reverse=reverse):
        if c[k] == 1:
            rv[k] = i
            i += 1
        else:
            v = c[k]
            rv[k] = (2*i + v - 1)/2
            i += v
    return np.array([rv[k] for k in a], dtype='float')


def rrankdata(a):
    """Alias to `rankdata(a, reverse=True)`. See `rankdata` for details."""
    return rankdata(a, reverse=True)



import numpy as np
from pymcdm.methods import TOPSIS
from pymcdm.helpers import rrankdata

# Define decision matrix (2 criteria, 4 alternative)
alts = np.array([
    [4, 4],
    [1, 5],
    [3, 2],
    [4, 2]
    ], dtype='float')

# Define weights and types
weights = np.array([0.5, 0.5])
types = np.array([1, -1])

# Create object of the method
topsis = TOPSIS()

# Determine preferences and ranking for alternatives
pref = topsis(alts, weights, types)
ranking = rrankdata(pref)

for r, p in zip(ranking, pref):
    print(r, p)



# Copyright (c) 2020 Andrii Shekhovtsov

import numpy as np
from .. import normalizations
from .mcda_method import MCDA_method

class SPOTIS(MCDA_method):
    def __init__(self):
        """Create SPOTIS method object."""
        pass

    def __call__(self, matrix, weights, types, bounds, *args, **kwargs):
        """Rank alternatives from decision matrix `matrix`, with criteria weights `weights` and criteria types `types`.

Parameters
----------
    matrix : ndarray
        Decision matrix / alternatives data.
        Alternatives are in rows and Criteria are in columns.

    weights : ndarray
        Criteria weights. Sum of the weights should be 1. (e.g. sum(weights) == 1)

    types : ndarray
        Array with definitions of criteria types:
        1 if criteria is profit and -1 if criteria is cost for each criteria in `matrix`.

    bounds : ndarray
        Each row should contain min and max values for each criterion. Min and max should be different values!

    *args and **kwargs are necessary for methods which reqiure some additional data.

Returns
-------
    ndarray
        Preference values for alternatives. Better alternatives have smaller values.
"""
        SPOTIS._validate_input_data(matrix, weights, types)
        if np.any(bounds[:, 0] == bounds[:, 1]):
            eq = np.arange(bounds.shape[0])[bounds[:, 0] == bounds[:, 1]]
            raise ValueError(
                    f'Bounds for criteria {eq} are equal. Consider changing min and max values for this criterion, delete this criterion or use another MCDA method.'
                )

        # Determine Ideal Solution Point based on criteria bounds
        isp = bounds[np.arange(bounds.shape[0]), ((types+1)//2).astype('int')]
        return SPOTIS._spotis(matrix, weights, isp, bounds)

    @staticmethod
    def _spotis(matrix, weights, isp, bounds):
        nmatrix = matrix.astype(float)
        # Normalized distances matrix (d_{ij})
        nmatrix = np.abs((nmatrix - isp)/
                         (bounds[:,0] - bounds[:,1]))
        # Distances to ISP (smaller means better alt)
        raw_scores = np.sum(nmatrix * weights, axis=1)
        return raw_scores



# Copyright (c) 2020 Andrii Shekhovtsov

from abc import ABC

class MCDA_method(ABC):
    def __call__(self, matrix, weights, types, *args, **kwargs):
        """Rank alternatives from decision matrix `matrix`, with criteria weights `weights` and criteria types `types`.

Parameters
----------
    matrix : ndarray
        Decision matrix / alternatives data.
        Alternatives are in rows and Criteria are in columns.

    weights : ndarray
        Criteria weights. Sum of the weights should be 1. (e.g. sum(weights) == 1)

    types : ndarray
        Array with definitions of criteria types:
        1 if criteria is profit and -1 if criteria is cost for each criteria in `matrix`.

    *args and **kwargs are necessary for methods which reqiure some additional data.
"""
        pass

    @staticmethod
    def _validate_input_data(matrix, weights, types):
        if matrix.shape[1] != weights.shape[0] and weights.shape[0] != len(types):
            raise ValueError(f'Number of criteria should be same as number of weights and number of types')



# Copyright (c) 2020 Andrii Shekhovtsov

import numpy as np
from itertools import permutations

__all__ = [
    'spearman',
    'pearson',
    'weighted_spearman',
    'rank_similarity_coef',
    'kendall_tau',
    'goodman_kruskal_gamma'
]

def _cov(x, y):
    return np.cov(x, y, bias=True)[0][1]

def spearman(x, y):
    """Calculate Spearman correlation between two rankings vectors.

Parameters
----------
    x : ndarray
        First vector of ranks.

    y : ndarray
        Second vector of ranks.

Returns
-------
    float
        Correlation between two rankings vectors.
"""
    return (_cov(x, y)) / (np.std(x) * np.std(y))


def pearson(x, y):
    """Calculate Pearson correlation between two raw vectors.

Parameters
----------
    x : ndarray
        First vector with raw values.

    y : ndarray
        Second vector with raw values.

Returns
-------
    float
        Correlation between two vectors.
"""
    return (_cov(x, y)) / (np.std(x) * np.std(y))


def weighted_spearman(x, y):
    """Calculate Weighted Spearman correlation between two rankings vectors.

Parameters
----------
    x : ndarray
        First vector of ranks.

    y : ndarray
        Second vector of ranks.

Returns
-------
    float
        Correlation between two rankings vectors.
"""
    N = len(x)
    n = 6 * np.sum((x-y)**2 * ((N - x + 1) + (N - y + 1)))
    d = N**4 + N**3 - N**2 - N
    return 1 - (n/d)


def rank_similarity_coef(x, y):
    """Calculate Rank Similarity Coefficient (WS) between two rankings vectors.

Parameters
----------
    x : ndarray
        First vector of ranks.

    y : ndarray
        Second vector of ranks.

Returns
-------
    float
        Correlation between two rankings vectors.
"""
    N = len(x)
    n = np.fabs(x - y)
    d = np.max((np.fabs(1 - x), np.fabs(N - x)), axis=0)
    return 1 - np.sum(2.0**(-1.0 * x) * n/d)


def kendall_tau(x, y):
    """Calculate Kendall Tau correlation between two rankings vectors.

Parameters
----------
    x : ndarray
        First vector of ranks.

    y : ndarray
        Second vector of ranks.

Returns
-------
    float
        Correlation between two rankings vectors.
"""
    n = len(x)
    res = 0
    for j in range(n):
        for i in range(j):
            res += np.sign(x[i] - x[j]) * np.sign(y[i] - y[j])
    return 2/(n*(n-1)) * res


def goodman_kruskal_gamma(x, y):
    """Calculate Goodman's and Kruskal's Gamma correlation between two rankings vectors.

Parameters
----------
    x : ndarray
        First vector of ranks.

    y : ndarray
        Second vector of ranks.

Returns
-------
    float
        Correlation between two rankings vectors.
"""
    num = 0
    den = 0
    for i, j in permutations(range(len(x)), 2):
        x_dir = x[i] - x[j]
        y_dir = y[i] - y[j]
        sign = np.sign(x_dir * y_dir)
        num += sign
        if sign:
            den += 1
    return num / float(den)



# Copyright (c) 2020 Andrii Shekhovtsov

import numpy as np
from .normalizations import sum_normalization, normalize_matrix

__all__ = [
    'equal',
    'entropy',
    'standard_deviation',
]

def equal(matrix):
    """Calculate equal weights for given `matrix`.

Parameters
----------
    matrix : ndarray
        Decision matrix / alternatives data.
        Alternatives are in rows and Criteria are in columns.

Returns
-------
    ndarray
        Vector of weights.
"""
    N = matrix.shape[1]
    return np.ones(N) / N

def entropy(matrix):
    """Calculate weights for given `matrix` using entropy method.

Parameters
----------
    matrix : ndarray
        Decision matrix / alternatives data.
        Alternatives are in rows and Criteria are in columns.

Returns
-------
    ndarray
        Vector of weights.
"""
    m, n = matrix.shape
    nmatrix = normalize_matrix(matrix, sum_normalization, None)
    entropies = np.empty(n)
    # Iterate over all criteria
    for i, col in enumerate(nmatrix.T):
        if np.any(col == 0):
            entropies[i] = 0
        else:
            entropies[i] = -np.sum(col * np.log(col))
    entropies = entropies / np.log(m)

    E = 1 - entropies
    return E / np.sum(E)


def standard_deviation(matrix):
    """Calculate weights for given `matrix` using std method.

Parameters
----------
    matrix : ndarray
        Decision matrix / alternatives data.
        Alternatives are in rows and Criteria are in columns.

Returns
-------
    ndarray
        Vector of weights.
"""
    m, n = matrix.shape
    std = np.std(matrix, axis=0)
    return std / np.sum(std)



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pymcdm",
    version="1.0.3.post2",
    author="Andrii Shekhovtsov",
    author_email="andrii-shekhovtsov@zut.edu.pl",
    description="Python library for Multi-Criteria Decision-Making",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://gitlab.com/shekhand/mcda",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.7',
    install_requires=[
        'numpy',
    ]
)

